/* -----------------------------------------
  countedptr.hpp
  
  All the code in this file is open source
  so feel free to use it any way you would
  like.
  
  designed and implemented by
  ArchMiffo, member of Darkbits
----------------------------------------- */

#ifndef AUL_COUNTED_PTR_HPP
#define AUL_COUNTED_PTR_HPP

// aul stands for ArchMiffo Utility Library
namespace aul
{

/* ------------------------------------------------------
  The counted_ptr class

  Acts like a real pointer but it keeps track of its
  pointee and deletes it when the last object that uses 
  it goes out of scope. This is a very simple type of
  smart pointer but it does what it's supposed to do. 
  Unlike the standard auto_ptr class you can use this 
  pointer class in the standard containers (vector,
  list, etc) and you don't have to remember to free
  the memory when you are done with the pointer.
------------------------------------------------------ */
template<typename T>
class counted_ptr
{
public:  
  // Some typedefs for accessing the
  // type of the pointer
  typedef T value_type;
  typedef T* pointer;
  typedef T& reference;

  explicit counted_ptr(T* ptr=0) 
           : pointee_(ptr), refCount_(new int(1))
           {}
  
  ~counted_ptr() { decrease(); }

  counted_ptr(const counted_ptr<T>& cp)
  : pointee_(cp.pointee_), refCount_(cp.refCount_)
  { ++*refCount_; }

  // This constructor is used for implicit conversion
  // between diffrent types of counted_ptr  
  template<typename U>
  counted_ptr(const counted_ptr<U>& cp)
  : pointee_(cp.pointee_), refCount_(cp.refCount_)
  { ++*refCount_; }

  counted_ptr<T>& operator=(const counted_ptr<T>& cp);

  // Conversion to regular pointer
  T* toPointer() { return pointee_; }
  
  T& operator*() { return *pointee_; }
  T* operator->() { return pointee_; }

  // To check for null pointer
  bool operator!() { return !pointee_; }

  int getCount() { return *refCount_; }

  T* extractPointer();

private:
  // Decrease refcount and clean 
  // up if this is the last object
  void decrease()
  {
    if(--*refCount_ == 0) 
    { 
      delete pointee_;
      delete refCount_;
    } // end if 
  } // end decrease
  
  template<typename U> 
  friend class counted_ptr;

  T* pointee_;
  int* refCount_;
}; // end counted_ptr

/* ----------------
  Copy assignment
---------------- */
template<typename T>
counted_ptr<T>& counted_ptr<T>::operator=(const counted_ptr<T>& cp)
{
  if(pointee_ == cp.pointee_)
    return *this;

  decrease();

  pointee_ = cp.pointee_;
  refCount_ = cp.refCount_;
  
  ++*refCount_;

  return *this;
} // end counted_ptr<T>::operator=

/* --------------------------------------------------
  extractPointer

  Set the pointer to 0.
  After this the client is responsible for deleting
  the returned pointer
-------------------------------------------------- */
template<typename T>
T* counted_ptr<T>::extractPointer()
{
  T* ptr = pointee_;

  pointee_ = 0;

  return ptr;
} // end counted_ptr<T>::extractPointer

} // en namespace aul

#endif // end AUL_COUNTED_PTR_HPP